# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class EpicsConnectionInfo(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsEpicsConnectionInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = EpicsConnectionInfo()
        x.Init(buf, n + offset)
        return x

    # EpicsConnectionInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # EpicsConnectionInfo
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EpicsConnectionInfo
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # EpicsConnectionInfo
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # EpicsConnectionInfo
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def EpicsConnectionInfoStart(builder):
    builder.StartObject(4)


def EpicsConnectionInfoAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(0, timestamp, 0)


def EpicsConnectionInfoAddType(builder, type):
    builder.PrependUint16Slot(1, type, 0)


def EpicsConnectionInfoAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def EpicsConnectionInfoAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def EpicsConnectionInfoEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class SpectraDetectorMapping(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsSpectraDetectorMapping(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SpectraDetectorMapping()
        x.Init(buf, n + offset)
        return x

    # SpectraDetectorMapping
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SpectraDetectorMapping
    def Spectrum(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Int32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # SpectraDetectorMapping
    def SpectrumAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SpectraDetectorMapping
    def SpectrumLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SpectraDetectorMapping
    def DetectorId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Int32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # SpectraDetectorMapping
    def DetectorIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SpectraDetectorMapping
    def DetectorIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SpectraDetectorMapping
    def NSpectra(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0


def SpectraDetectorMappingStart(builder):
    builder.StartObject(3)


def SpectraDetectorMappingAddSpectrum(builder, spectrum):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(spectrum), 0
    )


def SpectraDetectorMappingStartSpectrumVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def SpectraDetectorMappingAddDetectorId(builder, detectorId):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(detectorId), 0
    )


def SpectraDetectorMappingStartDetectorIdVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def SpectraDetectorMappingAddNSpectra(builder, nSpectra):
    builder.PrependInt32Slot(2, nSpectra, 0)


def SpectraDetectorMappingEnd(builder):
    return builder.EndObject()



from collections import namedtuple

import flatbuffers

from streaming_data_types.fbschemas.nicos_cache_ns10 import CacheEntry
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"ns10"


def serialise_ns10(
    key: str, value: str, time_stamp: float = 0, ttl: float = 0, expired: bool = False
):
    builder = flatbuffers.Builder(128)
    builder.ForceDefaults(True)

    value_offset = builder.CreateString(value)
    key_offset = builder.CreateString(key)

    CacheEntry.CacheEntryStart(builder)
    CacheEntry.CacheEntryAddValue(builder, value_offset)
    CacheEntry.CacheEntryAddExpired(builder, expired)
    CacheEntry.CacheEntryAddTtl(builder, ttl)
    CacheEntry.CacheEntryAddTime(builder, time_stamp)
    CacheEntry.CacheEntryAddKey(builder, key_offset)
    cache_entry_message = CacheEntry.CacheEntryEnd(builder)

    builder.Finish(cache_entry_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


def deserialise_ns10(buffer):
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    entry = CacheEntry.CacheEntry.GetRootAsCacheEntry(buffer, 0)

    key = entry.Key() if entry.Key() else b""
    time_stamp = entry.Time()
    ttl = entry.Ttl() if entry.Ttl() else 0
    expired = entry.Expired() if entry.Expired() else False
    value = entry.Value() if entry.Value() else b""

    Entry = namedtuple("Entry", ("key", "time_stamp", "ttl", "expired", "value"))

    return Entry(key.decode().strip(), time_stamp, ttl, expired, value.decode())



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class SampleEnvironmentData(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsSampleEnvironmentData(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SampleEnvironmentData()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def SampleEnvironmentDataBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x73\x65\x6E\x76", size_prefixed=size_prefixed
        )

    # SampleEnvironmentData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SampleEnvironmentData
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # SampleEnvironmentData
    def Channel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # SampleEnvironmentData
    def PacketTimestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # SampleEnvironmentData
    def TimeDelta(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float64Flags, o + self._tab.Pos
            )
        return 0.0

    # SampleEnvironmentData
    def TimestampLocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # SampleEnvironmentData
    def ValuesType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # SampleEnvironmentData
    def Values(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # SampleEnvironmentData
    def Timestamps(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # SampleEnvironmentData
    def TimestampsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # SampleEnvironmentData
    def TimestampsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SampleEnvironmentData
    def TimestampsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # SampleEnvironmentData
    def MessageCounter(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0


def SampleEnvironmentDataStart(builder):
    builder.StartObject(9)


def SampleEnvironmentDataAddName(builder, Name):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(Name), 0
    )


def SampleEnvironmentDataAddChannel(builder, Channel):
    builder.PrependInt32Slot(1, Channel, 0)


def SampleEnvironmentDataAddPacketTimestamp(builder, PacketTimestamp):
    builder.PrependUint64Slot(2, PacketTimestamp, 0)


def SampleEnvironmentDataAddTimeDelta(builder, TimeDelta):
    builder.PrependFloat64Slot(3, TimeDelta, 0.0)


def SampleEnvironmentDataAddTimestampLocation(builder, TimestampLocation):
    builder.PrependInt8Slot(4, TimestampLocation, 0)


def SampleEnvironmentDataAddValuesType(builder, ValuesType):
    builder.PrependUint8Slot(5, ValuesType, 0)


def SampleEnvironmentDataAddValues(builder, Values):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(Values), 0
    )


def SampleEnvironmentDataAddTimestamps(builder, Timestamps):
    builder.PrependUOffsetTRelativeSlot(
        7, flatbuffers.number_types.UOffsetTFlags.py_type(Timestamps), 0
    )


def SampleEnvironmentDataStartTimestampsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def SampleEnvironmentDataAddMessageCounter(builder, MessageCounter):
    builder.PrependUint64Slot(8, MessageCounter, 0)


def SampleEnvironmentDataEnd(builder):
    return builder.EndObject()



import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.epics_connection_info_ep00 import (
    deserialise_ep00,
    serialise_ep00,
)
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.fbschemas.epics_connection_info_ep00 import EventType


class TestSerialisationEp00:
    original_entry = {
        "timestamp_ns": 1593620746000000000,
        "event_type": EventType.EventType.DISCONNECTED,
        "source_name": "test_source",
        "service_id": "test_service",
    }

    def test_serialises_and_deserialises_ep00_message_correctly(self):
        buf = serialise_ep00(**self.original_entry)
        deserialised_tuple = deserialise_ep00(buf)

        assert deserialised_tuple.timestamp == self.original_entry["timestamp_ns"]
        assert deserialised_tuple.type == self.original_entry["event_type"]
        assert deserialised_tuple.source_name == self.original_entry["source_name"]
        assert deserialised_tuple.service_id == self.original_entry["service_id"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_ep00(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_ep00(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "ep00" in SERIALISERS
        assert "ep00" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class ConfigUpdate(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsConfigUpdate(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ConfigUpdate()
        x.Init(buf, n + offset)
        return x

    # ConfigUpdate
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ConfigUpdate
    def ConfigChange(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # ConfigUpdate
    def Streams(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Stream import Stream

            obj = Stream()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ConfigUpdate
    def StreamsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0


def ConfigUpdateStart(builder):
    builder.StartObject(2)


def ConfigUpdateAddConfigChange(builder, configChange):
    builder.PrependUint16Slot(0, configChange, 0)


def ConfigUpdateAddStreams(builder, streams):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(streams), 0
    )


def ConfigUpdateStartStreamsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def ConfigUpdateEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class EventHistogram(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsEventHistogram(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = EventHistogram()
        x.Init(buf, n + offset)
        return x

    # EventHistogram
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # EventHistogram
    def Source(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # EventHistogram
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EventHistogram
    def DimMetadata(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .DimensionMetaData import DimensionMetaData

            obj = DimensionMetaData()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # EventHistogram
    def DimMetadataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EventHistogram
    def LastMetadataTimestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EventHistogram
    def CurrentShape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # EventHistogram
    def CurrentShapeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # EventHistogram
    def CurrentShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EventHistogram
    def Offset(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # EventHistogram
    def OffsetAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # EventHistogram
    def OffsetLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EventHistogram
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # EventHistogram
    def Data(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # EventHistogram
    def ErrorsType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # EventHistogram
    def Errors(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # EventHistogram
    def Info(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def EventHistogramStart(builder):
    builder.StartObject(11)


def EventHistogramAddSource(builder, source):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(source), 0
    )


def EventHistogramAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(1, timestamp, 0)


def EventHistogramAddDimMetadata(builder, dimMetadata):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(dimMetadata), 0
    )


def EventHistogramStartDimMetadataVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def EventHistogramAddLastMetadataTimestamp(builder, lastMetadataTimestamp):
    builder.PrependUint64Slot(3, lastMetadataTimestamp, 0)


def EventHistogramAddCurrentShape(builder, currentShape):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(currentShape), 0
    )


def EventHistogramStartCurrentShapeVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def EventHistogramAddOffset(builder, offset):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(offset), 0
    )


def EventHistogramStartOffsetVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def EventHistogramAddDataType(builder, dataType):
    builder.PrependUint8Slot(6, dataType, 0)


def EventHistogramAddData(builder, data):
    builder.PrependUOffsetTRelativeSlot(
        7, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0
    )


def EventHistogramAddErrorsType(builder, errorsType):
    builder.PrependUint8Slot(8, errorsType, 0)


def EventHistogramAddErrors(builder, errors):
    builder.PrependUOffsetTRelativeSlot(
        9, flatbuffers.number_types.UOffsetTFlags.py_type(errors), 0
    )


def EventHistogramAddInfo(builder, info):
    builder.PrependUOffsetTRelativeSlot(
        10, flatbuffers.number_types.UOffsetTFlags.py_type(info), 0
    )


def EventHistogramEnd(builder):
    return builder.EndObject()



from datetime import datetime, timezone
from struct import pack
from typing import List, NamedTuple, Union

import flatbuffers
import numpy as np

import streaming_data_types.fbschemas.ADAr_ADArray_schema.Attribute as ADArAttribute
from streaming_data_types.fbschemas.ADAr_ADArray_schema import ADArray
from streaming_data_types.fbschemas.ADAr_ADArray_schema.DType import DType
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"ADAr"


class Attribute:
    def __init__(
        self,
        name: str,
        description: str,
        source: str,
        data: Union[np.ndarray, str, int, float],
    ):
        self.name = name
        self.description = description
        self.source = source
        self.data = data

    def __eq__(self, other):
        data_is_equal = type(self.data) == type(other.data)  # noqa: E721
        if type(self.data) is np.ndarray:
            data_is_equal = data_is_equal and np.array_equal(self.data, other.data)
        else:
            data_is_equal = data_is_equal and self.data == other.data
        return (
            self.name == other.name
            and self.description == other.description
            and self.source == other.source
            and data_is_equal
        )


def serialise_ADAr(
    source_name: str,
    unique_id: int,
    timestamp: datetime,
    data: Union[np.ndarray, str],
    attributes: List[Attribute] = [],
) -> bytes:
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    type_map = {
        np.dtype("uint8"): DType.uint8,
        np.dtype("int8"): DType.int8,
        np.dtype("uint16"): DType.uint16,
        np.dtype("int16"): DType.int16,
        np.dtype("uint32"): DType.uint32,
        np.dtype("int32"): DType.int32,
        np.dtype("uint64"): DType.uint64,
        np.dtype("int64"): DType.int64,
        np.dtype("float32"): DType.float32,
        np.dtype("float64"): DType.float64,
    }

    if type(data) is str:
        data = np.frombuffer(data.encode(), np.uint8)
        data_type = DType.c_string
    else:
        data_type = type_map[data.dtype]

    # Build dims
    dims_offset = builder.CreateNumpyVector(np.asarray(data.shape))

    # Build data
    data_offset = builder.CreateNumpyVector(data.flatten().view(np.uint8))

    source_name_offset = builder.CreateString(source_name)

    temp_attributes = []
    for item in attributes:
        if type(item.data) is np.ndarray:
            attr_data_type = type_map[item.data.dtype]
            attr_data = item.data
        elif type(item.data) is str:
            attr_data_type = DType.c_string
            attr_data = np.frombuffer(item.data.encode(), np.uint8)
        elif type(item.data) is int:
            attr_data_type = DType.int64
            attr_data = np.frombuffer(pack("q", item.data), np.uint8)
        elif type(item.data) is float:
            attr_data_type = DType.float64
            attr_data = np.frombuffer(pack("d", item.data), np.uint8)
        attr_name_offset = builder.CreateString(item.name)
        attr_desc_offset = builder.CreateString(item.description)
        attr_src_offset = builder.CreateString(item.source)
        attr_data_offset = builder.CreateNumpyVector(attr_data.flatten().view(np.uint8))
        ADArAttribute.AttributeStart(builder)
        ADArAttribute.AttributeAddName(builder, attr_name_offset)
        ADArAttribute.AttributeAddDescription(builder, attr_desc_offset)
        ADArAttribute.AttributeAddSource(builder, attr_src_offset)
        ADArAttribute.AttributeAddDataType(builder, attr_data_type)
        ADArAttribute.AttributeAddData(builder, attr_data_offset)
        attr_offset = ADArAttribute.AttributeEnd(builder)
        temp_attributes.append(attr_offset)

    ADArray.ADArrayStartAttributesVector(builder, len(attributes))
    for item in reversed(temp_attributes):
        builder.PrependUOffsetTRelative(item)
    attributes_offset = builder.EndVector(len(attributes))

    # Build the actual buffer
    ADArray.ADArrayStart(builder)
    ADArray.ADArrayAddSourceName(builder, source_name_offset)
    ADArray.ADArrayAddDataType(builder, data_type)
    ADArray.ADArrayAddDimensions(builder, dims_offset)
    ADArray.ADArrayAddId(builder, unique_id)
    ADArray.ADArrayAddData(builder, data_offset)
    ADArray.ADArrayAddTimestamp(builder, int(timestamp.timestamp() * 1e9))
    ADArray.ADArrayAddAttributes(builder, attributes_offset)
    array_message = ADArray.ADArrayEnd(builder)

    builder.Finish(array_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


ADArray_t = NamedTuple(
    "ADArray",
    (
        ("source_name", str),
        ("unique_id", int),
        ("timestamp", datetime),
        ("data", np.ndarray),
        ("attributes", List[Attribute]),
    ),
)


def get_payload_data(fb_arr) -> np.ndarray:
    return get_data(fb_arr).reshape(fb_arr.DimensionsAsNumpy())


def get_data(fb_arr) -> np.ndarray:
    """
    Converts the data array into the correct type.
    """
    raw_data = fb_arr.DataAsNumpy()
    type_map = {
        DType.uint8: np.uint8,
        DType.int8: np.int8,
        DType.uint16: np.uint16,
        DType.int16: np.int16,
        DType.uint32: np.uint32,
        DType.int32: np.int32,
        DType.uint64: np.uint64,
        DType.int64: np.int64,
        DType.float32: np.float32,
        DType.float64: np.float64,
    }
    return raw_data.view(type_map[fb_arr.DataType()])


def deserialise_ADAr(buffer: Union[bytearray, bytes]) -> ADArray:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    ad_array = ADArray.ADArray.GetRootAsADArray(buffer, 0)
    unique_id = ad_array.Id()
    max_time = datetime(
        year=9000, month=1, day=1, hour=0, minute=0, second=0
    ).timestamp()
    used_timestamp = ad_array.Timestamp() / 1e9
    if used_timestamp > max_time:
        used_timestamp = max_time
    if ad_array.DataType() == DType.c_string:
        data = ad_array.DataAsNumpy().tobytes().decode()
    else:
        data = get_payload_data(ad_array)

    attributes_list = []
    for i in range(ad_array.AttributesLength()):
        attribute_ptr = ad_array.Attributes(i)
        if attribute_ptr.DataType() == DType.c_string:
            attr_data = attribute_ptr.DataAsNumpy().tobytes().decode()
        else:
            attr_data = get_data(attribute_ptr)
        temp_attribute = Attribute(
            name=attribute_ptr.Name().decode(),
            description=attribute_ptr.Description().decode(),
            source=attribute_ptr.Source().decode(),
            data=attr_data,
        )
        if type(temp_attribute.data) is np.ndarray and len(temp_attribute.data) == 1:
            if np.issubdtype(temp_attribute.data.dtype, np.floating):
                temp_attribute.data = float(temp_attribute.data[0])
            elif np.issubdtype(temp_attribute.data.dtype, np.integer):
                temp_attribute.data = int(temp_attribute.data[0])
        attributes_list.append(temp_attribute)

    return ADArray_t(
        source_name=ad_array.SourceName().decode(),
        unique_id=unique_id,
        timestamp=datetime.fromtimestamp(used_timestamp, tz=timezone.utc),
        data=data,
        attributes=attributes_list,
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Int64(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Int64()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsInt64(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def Int64BufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x73\x63\x61\x6C", size_prefixed=size_prefixed
        )

    # Int64
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Int64
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0


def Int64Start(builder):
    builder.StartObject(1)


def Start(builder):
    return Int64Start(builder)


def Int64AddValue(builder, value):
    builder.PrependInt64Slot(0, value, 0)


def AddValue(builder, value):
    return Int64AddValue(builder, value)


def Int64End(builder):
    return builder.EndObject()


def End(builder):
    return Int64End(builder)



from collections import namedtuple

import flatbuffers

from streaming_data_types.fbschemas.status_x5f2 import Status
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"x5f2"

StatusMessage = namedtuple(
    "StatusMessage",
    (
        "software_name",
        "software_version",
        "service_id",
        "host_name",
        "process_id",
        "update_interval",
        "status_json",
    ),
)


def deserialise_x5f2(buffer):
    """
    Deserialise FlatBuffer x5f2.

    :param buffer: The FlatBuffers buffer.
    :return: The deserialised data.
    """
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    log_message = Status.Status.GetRootAsStatus(buffer, 0)

    return StatusMessage(
        log_message.SoftwareName().decode("utf-8"),
        log_message.SoftwareVersion().decode("utf-8"),
        log_message.ServiceId().decode("utf-8"),
        log_message.HostName().decode("utf-8"),
        log_message.ProcessId(),
        log_message.UpdateInterval(),
        log_message.StatusJson().decode("utf-8"),
    )


def serialise_x5f2(
    software_name: str,
    software_version: str,
    service_id: str,
    host_name: str,
    process_id: int,
    update_interval: int,
    status_json: str,
) -> bytes:
    """
    Serialise status message as an x5f2 FlatBuffers message.

    :param software_name:
    :param software_version:
    :param service_id:
    :param host_name:
    :param process_id:
    :param update_interval:
    :param status_json:
    :return:
    """

    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    software_name = builder.CreateString(software_name)
    software_version = builder.CreateString(software_version)
    service_id = builder.CreateString(service_id)
    host_name = builder.CreateString(host_name)
    status_json = builder.CreateString(status_json)

    # Build the actual buffer
    Status.StatusStart(builder)

    Status.StatusAddSoftwareName(builder, software_name)
    Status.StatusAddSoftwareVersion(builder, software_version)
    Status.StatusAddServiceId(builder, service_id)
    Status.StatusAddHostName(builder, host_name)
    Status.StatusAddProcessId(builder, process_id)
    Status.StatusAddUpdateInterval(builder, update_interval)
    Status.StatusAddStatusJson(builder, status_json)

    data = Status.StatusEnd(builder)
    builder.Finish(data, file_identifier=FILE_IDENTIFIER)

    return bytes(builder.Output())



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class ValueUnion(object):
    NONE = 0
    Int8Array = 1
    UInt8Array = 2
    Int16Array = 3
    UInt16Array = 4
    Int32Array = 5
    UInt32Array = 6
    Int64Array = 7
    UInt64Array = 8



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Status(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsStatus(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Status()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def StatusBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x78\x35\x66\x32", size_prefixed=size_prefixed
        )

    # Status
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Status
    def SoftwareName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def SoftwareVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def HostName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Status
    def ProcessId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # Status
    def UpdateInterval(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # Status
    def StatusJson(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def StatusStart(builder):
    builder.StartObject(7)


def StatusAddSoftwareName(builder, softwareName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(softwareName), 0
    )


def StatusAddSoftwareVersion(builder, softwareVersion):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(softwareVersion), 0
    )


def StatusAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def StatusAddHostName(builder, hostName):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(hostName), 0
    )


def StatusAddProcessId(builder, processId):
    builder.PrependUint32Slot(4, processId, 0)


def StatusAddUpdateInterval(builder, updateInterval):
    builder.PrependUint32Slot(5, updateInterval, 0)


def StatusAddStatusJson(builder, statusJson):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(statusJson), 0
    )


def StatusEnd(builder):
    return builder.EndObject()



import pytest
from datetime import datetime, timezone
from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types import (
    deserialise_pvAl,
    serialise_pvAl,
)
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.epics_pv_alarm_status_pvAl import (
    AlarmState,
    CAAlarmState,
    AlarmSeverity,
)


class TestSerialisation_pvAl:
    original_entry = {
        "source_name": "test_source",
        "timestamp": datetime(
            year=2021,
            month=3,
            day=26,
            hour=12,
            minute=32,
            second=11,
            tzinfo=timezone.utc,
        ),
        "state": AlarmState.CLIENT,
        "ca_state": CAAlarmState.BAD_SUB,
        "severity": AlarmSeverity.MAJOR,
        "message": "some_msg",
    }

    def test_serialises_and_deserialises_pvAl_message(self):
        buf = serialise_pvAl(**self.original_entry)
        deserialised_tuple = deserialise_pvAl(buf)

        assert deserialised_tuple.timestamp == self.original_entry["timestamp"]
        assert deserialised_tuple.state == self.original_entry["state"]
        assert deserialised_tuple.source_name == self.original_entry["source_name"]
        assert deserialised_tuple.ca_state == self.original_entry["ca_state"]
        assert deserialised_tuple.severity == self.original_entry["severity"]
        assert deserialised_tuple.message == self.original_entry["message"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_pvAl(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_pvAl(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "pvAl" in SERIALISERS
        assert "pvAl" in DESERIALISERS



from collections import namedtuple

import flatbuffers
import numpy as np

import streaming_data_types.fbschemas.eventdata_ev42.EventMessage as EventMessage
import streaming_data_types.fbschemas.eventdata_ev42.FacilityData as FacilityData
import streaming_data_types.fbschemas.isis_event_info_is84.ISISData as ISISData
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"ev42"


EventData = namedtuple(
    "EventData",
    (
        "source_name",
        "message_id",
        "pulse_time",
        "time_of_flight",
        "detector_id",
        "specific_data",
    ),
)


def deserialise_ev42(buffer):
    """
    Deserialise FlatBuffer ev42.

    :param buffer: The FlatBuffers buffer.
    :return: The deserialised data.
    """
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    event = EventMessage.EventMessage.GetRootAsEventMessage(buffer, 0)

    specific_data = None
    if event.FacilitySpecificDataType() == FacilityData.FacilityData.ISISData:
        specific = event.FacilitySpecificData()
        isis_buf = ISISData.ISISData()
        isis_buf.Init(specific.Bytes, specific.Pos)
        specific_data = {
            "period_number": isis_buf.PeriodNumber(),
            "run_state": isis_buf.RunState(),
            "proton_charge": isis_buf.ProtonCharge(),
        }

    return EventData(
        event.SourceName().decode("utf-8"),
        event.MessageId(),
        event.PulseTime(),
        event.TimeOfFlightAsNumpy(),
        event.DetectorIdAsNumpy(),
        specific_data,
    )


def serialise_ev42(
    source_name, message_id, pulse_time, time_of_flight, detector_id, isis_specific=None
):
    """
    Serialise event data as an ev42 FlatBuffers message.

    :param source_name:
    :param message_id:
    :param pulse_time:
    :param time_of_flight:
    :param detector_id:
    :param isis_specific:
    :return:
    """
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    source = builder.CreateString(source_name)

    tof_data = builder.CreateNumpyVector(np.asarray(time_of_flight).astype(np.uint32))
    det_data = builder.CreateNumpyVector(np.asarray(detector_id).astype(np.uint32))

    isis_data = None
    if isis_specific:
        # isis_builder = flatbuffers.Builder(96)
        ISISData.ISISDataStart(builder)
        ISISData.ISISDataAddPeriodNumber(builder, isis_specific["period_number"])
        ISISData.ISISDataAddRunState(builder, isis_specific["run_state"])
        ISISData.ISISDataAddProtonCharge(builder, isis_specific["proton_charge"])
        isis_data = ISISData.ISISDataEnd(builder)

    # Build the actual buffer
    EventMessage.EventMessageStart(builder)
    EventMessage.EventMessageAddDetectorId(builder, det_data)
    EventMessage.EventMessageAddTimeOfFlight(builder, tof_data)
    EventMessage.EventMessageAddPulseTime(builder, pulse_time)
    EventMessage.EventMessageAddMessageId(builder, message_id)
    EventMessage.EventMessageAddSourceName(builder, source)

    if isis_specific:
        EventMessage.EventMessageAddFacilitySpecificDataType(
            builder, FacilityData.FacilityData.ISISData
        )
        EventMessage.EventMessageAddFacilitySpecificData(builder, isis_data)

    data = EventMessage.EventMessageEnd(builder)

    builder.Finish(data, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())



from typing import Union, NamedTuple
from datetime import datetime, timezone

import flatbuffers

from streaming_data_types.fbschemas.pv_alarm_state_pvAl.AlarmState import AlarmState
from streaming_data_types.fbschemas.pv_alarm_state_pvAl.AlarmSeverity import (
    AlarmSeverity,
)
from streaming_data_types.fbschemas.pv_alarm_state_pvAl.CAAlarmState import CAAlarmState
from streaming_data_types.fbschemas.pv_alarm_state_pvAl import PV_AlarmState
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"pvAl"


def serialise_pvAl(
    source_name: str,
    timestamp: datetime,
    state: AlarmState,
    severity: AlarmSeverity,
    ca_state: CAAlarmState = CAAlarmState.NO_ALARM,
    message: str = "",
) -> bytes:
    builder = flatbuffers.Builder(136)
    builder.ForceDefaults(True)

    source_name_offset = builder.CreateString(source_name)
    message_offset = builder.CreateString(message)

    PV_AlarmState.PV_AlarmStateStart(builder)
    PV_AlarmState.PV_AlarmStateAddSourceName(builder, source_name_offset)
    PV_AlarmState.PV_AlarmStateAddMessage(builder, message_offset)
    PV_AlarmState.PV_AlarmStateAddState(builder, state)
    PV_AlarmState.PV_AlarmStateAddCaState(builder, ca_state)
    PV_AlarmState.PV_AlarmStateAddSeverity(builder, severity)
    PV_AlarmState.PV_AlarmStateAddTimestamp(builder, int(timestamp.timestamp() * 1e6))

    end = PV_AlarmState.PV_AlarmStateEnd(builder)
    builder.Finish(end, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


PV_AlarmStateInfo = NamedTuple(
    "PV_AlarmStateInfo",
    (
        ("source_name", str),
        ("timestamp", datetime),
        ("state", AlarmState),
        ("ca_state", CAAlarmState),
        ("severity", AlarmSeverity),
        ("message", str),
    ),
)


def deserialise_pvAl(buffer: Union[bytearray, bytes]) -> PV_AlarmStateInfo:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    epics_alarm_state = PV_AlarmState.PV_AlarmState.GetRootAs(buffer, 0)

    source_name = (
        epics_alarm_state.SourceName() if epics_alarm_state.SourceName() else b""
    )
    message = epics_alarm_state.Message() if epics_alarm_state.Message() else b""

    return PV_AlarmStateInfo(
        source_name=source_name.decode("utf-8"),
        timestamp=datetime.fromtimestamp(
            epics_alarm_state.Timestamp() / 1e6, tz=timezone.utc
        ),
        state=epics_alarm_state.State(),
        ca_state=epics_alarm_state.CaState(),
        severity=epics_alarm_state.Severity(),
        message=message.decode("utf-8"),
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class EventType(object):
    UNKNOWN = 0
    NEVER_CONNECTED = 1
    CONNECTED = 2
    DISCONNECTED = 3
    DESTROYED = 4



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class LogData(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsLogData(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = LogData()
        x.Init(buf, n + offset)
        return x

    # LogData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # LogData
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # LogData
    def ValueType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # LogData
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # LogData
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # LogData
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 22

    # LogData
    def Severity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 4


def LogDataStart(builder):
    builder.StartObject(6)


def LogDataAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def LogDataAddValueType(builder, valueType):
    builder.PrependUint8Slot(1, valueType, 0)


def LogDataAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def LogDataAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(3, timestamp, 0)


def LogDataAddStatus(builder, status):
    builder.PrependUint16Slot(4, status, 22)


def LogDataAddSeverity(builder, severity):
    builder.PrependUint16Slot(5, severity, 4)


def LogDataEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ActionResponse(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsActionResponse(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ActionResponse()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def ActionResponseBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x61\x6E\x73\x77", size_prefixed=size_prefixed
        )

    # ActionResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ActionResponse
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ActionResponse
    def JobId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ActionResponse
    def Action(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # ActionResponse
    def Outcome(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # ActionResponse
    def StatusCode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ActionResponse
    def StopTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # ActionResponse
    def Message(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ActionResponse
    def CommandId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def ActionResponseStart(builder):
    builder.StartObject(8)


def ActionResponseAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def ActionResponseAddJobId(builder, jobId):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(jobId), 0
    )


def ActionResponseAddAction(builder, action):
    builder.PrependInt8Slot(2, action, 0)


def ActionResponseAddOutcome(builder, outcome):
    builder.PrependInt8Slot(3, outcome, 0)


def ActionResponseAddStatusCode(builder, statusCode):
    builder.PrependInt32Slot(4, statusCode, 0)


def ActionResponseAddStopTime(builder, stopTime):
    builder.PrependUint64Slot(5, stopTime, 0)


def ActionResponseAddMessage(builder, message):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(message), 0
    )


def ActionResponseAddCommandId(builder, commandId):
    builder.PrependUOffsetTRelativeSlot(
        7, flatbuffers.number_types.UOffsetTFlags.py_type(commandId), 0
    )


def ActionResponseEnd(builder):
    return builder.EndObject()



from collections import namedtuple
from typing import Any, Callable, Dict, Tuple, Union

import flatbuffers
import numpy as np

from streaming_data_types.fbschemas.logdata_f142 import LogData
from streaming_data_types.fbschemas.logdata_f142.ArrayByte import (
    ArrayByte,
    ArrayByteAddValue,
    ArrayByteEnd,
    ArrayByteStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayDouble import (
    ArrayDouble,
    ArrayDoubleAddValue,
    ArrayDoubleEnd,
    ArrayDoubleStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayFloat import (
    ArrayFloat,
    ArrayFloatAddValue,
    ArrayFloatEnd,
    ArrayFloatStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayInt import (
    ArrayInt,
    ArrayIntAddValue,
    ArrayIntEnd,
    ArrayIntStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayLong import (
    ArrayLong,
    ArrayLongAddValue,
    ArrayLongEnd,
    ArrayLongStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayShort import (
    ArrayShort,
    ArrayShortAddValue,
    ArrayShortEnd,
    ArrayShortStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayString import (
    ArrayString,
    ArrayStringAddValue,
    ArrayStringEnd,
    ArrayStringStart,
    ArrayStringStartValueVector,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayUByte import (
    ArrayUByte,
    ArrayUByteAddValue,
    ArrayUByteEnd,
    ArrayUByteStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayUInt import (
    ArrayUInt,
    ArrayUIntAddValue,
    ArrayUIntEnd,
    ArrayUIntStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayULong import (
    ArrayULong,
    ArrayULongAddValue,
    ArrayULongEnd,
    ArrayULongStart,
)
from streaming_data_types.fbschemas.logdata_f142.ArrayUShort import (
    ArrayUShort,
    ArrayUShortAddValue,
    ArrayUShortEnd,
    ArrayUShortStart,
)
from streaming_data_types.fbschemas.logdata_f142.Byte import (
    Byte,
    ByteAddValue,
    ByteEnd,
    ByteStart,
)
from streaming_data_types.fbschemas.logdata_f142.Double import (
    Double,
    DoubleAddValue,
    DoubleEnd,
    DoubleStart,
)
from streaming_data_types.fbschemas.logdata_f142.Float import (
    Float,
    FloatAddValue,
    FloatEnd,
    FloatStart,
)
from streaming_data_types.fbschemas.logdata_f142.Int import (
    Int,
    IntAddValue,
    IntEnd,
    IntStart,
)
from streaming_data_types.fbschemas.logdata_f142.Long import (
    Long,
    LongAddValue,
    LongEnd,
    LongStart,
)
from streaming_data_types.fbschemas.logdata_f142.Short import (
    Short,
    ShortAddValue,
    ShortEnd,
    ShortStart,
)
from streaming_data_types.fbschemas.logdata_f142.String import (
    String,
    StringAddValue,
    StringEnd,
    StringStart,
)
from streaming_data_types.fbschemas.logdata_f142.UByte import (
    UByte,
    UByteAddValue,
    UByteEnd,
    UByteStart,
)
from streaming_data_types.fbschemas.logdata_f142.UInt import (
    UInt,
    UIntAddValue,
    UIntEnd,
    UIntStart,
)
from streaming_data_types.fbschemas.logdata_f142.ULong import (
    ULong,
    ULongAddValue,
    ULongEnd,
    ULongStart,
)
from streaming_data_types.fbschemas.logdata_f142.UShort import (
    UShort,
    UShortAddValue,
    UShortEnd,
    UShortStart,
)
from streaming_data_types.fbschemas.logdata_f142.Value import Value
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"f142"


def _complete_buffer(
    builder,
    timestamp_unix_ns: int,
    alarm_status: Union[int, None] = None,
    alarm_severity: Union[int, None] = None,
) -> bytearray:
    LogData.LogDataAddTimestamp(builder, timestamp_unix_ns)

    if alarm_status is not None:
        LogData.LogDataAddStatus(builder, alarm_status)
        # Only include severity if status was provided, it would be meaningless by itself
        if alarm_severity is not None:
            LogData.LogDataAddSeverity(builder, alarm_severity)

    log_msg = LogData.LogDataEnd(builder)

    builder.Finish(log_msg, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


def _setup_builder(source_name: str) -> Tuple[flatbuffers.Builder, int]:
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)
    source = builder.CreateString(source_name)
    return builder, source


def _serialise_byte(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    ByteStart(builder)
    ByteAddValue(builder, data.item())
    value_position = ByteEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.Byte)


def _serialise_bytearray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayByteStart(builder)
    ArrayByteAddValue(builder, array_offset)
    value_position = ArrayByteEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayByte)


def _serialise_ubyte(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    UByteStart(builder)
    UByteAddValue(builder, data.item())
    value_position = UByteEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.UByte)


def _serialise_ubytearray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayUByteStart(builder)
    ArrayUByteAddValue(builder, array_offset)
    value_position = ArrayUByteEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayUByte)


def _serialise_short(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    ShortStart(builder)
    ShortAddValue(builder, data.item())
    value_position = ShortEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.Short)


def _serialise_shortarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayShortStart(builder)
    ArrayShortAddValue(builder, array_offset)
    value_position = ArrayShortEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayShort)


def _serialise_ushort(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    UShortStart(builder)
    UShortAddValue(builder, data.item())
    value_position = UShortEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.UShort)


def _serialise_ushortarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayUShortStart(builder)
    ArrayUShortAddValue(builder, array_offset)
    value_position = ArrayUShortEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayUShort)


def _serialise_int(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    IntStart(builder)
    IntAddValue(builder, data.item())
    value_position = IntEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.Int)


def _serialise_intarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayIntStart(builder)
    ArrayIntAddValue(builder, array_offset)
    value_position = ArrayIntEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayInt)


def _serialise_uint(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    UIntStart(builder)
    UIntAddValue(builder, data.item())
    value_position = UIntEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.UInt)


def _serialise_uintarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayUIntStart(builder)
    ArrayUIntAddValue(builder, array_offset)
    value_position = ArrayUIntEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayUInt)


def _serialise_long(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    LongStart(builder)
    LongAddValue(builder, data.item())
    value_position = LongEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.Long)


def _serialise_longarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayLongStart(builder)
    ArrayLongAddValue(builder, array_offset)
    value_position = ArrayLongEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayLong)


def _serialise_ulong(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    ULongStart(builder)
    ULongAddValue(builder, data.item())
    value_position = ULongEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ULong)


def _serialise_ulongarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayULongStart(builder)
    ArrayULongAddValue(builder, array_offset)
    value_position = ArrayULongEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayULong)


def _serialise_float(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    FloatStart(builder)
    FloatAddValue(builder, data.item())
    value_position = FloatEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.Float)


def _serialise_floatarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayFloatStart(builder)
    ArrayFloatAddValue(builder, array_offset)
    value_position = ArrayFloatEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayFloat)


def _serialise_double(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    DoubleStart(builder)
    DoubleAddValue(builder, data.item())
    value_position = DoubleEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.Double)


def _serialise_doublearray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    array_offset = builder.CreateNumpyVector(data)
    ArrayDoubleStart(builder)
    ArrayDoubleAddValue(builder, array_offset)
    value_position = ArrayDoubleEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayDouble)


def _serialise_string(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    string_offset = builder.CreateString(data.item())
    StringStart(builder)
    StringAddValue(builder, string_offset)
    value_position = StringEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.String)


def _serialise_stringarray(builder: flatbuffers.Builder, data: np.ndarray, source: int):
    string_offsets = [
        builder.CreateString(string_item) for string_item in reversed(data)
    ]
    ArrayStringStartValueVector(builder, len(data))
    for string_offset in string_offsets:
        builder.PrependSOffsetTRelative(string_offset)
    string_array_offset = builder.EndVector(len(data))
    ArrayStringStart(builder)
    ArrayStringAddValue(builder, string_array_offset)
    value_position = ArrayStringEnd(builder)
    LogData.LogDataStart(builder)
    LogData.LogDataAddSourceName(builder, source)
    LogData.LogDataAddValue(builder, value_position)
    LogData.LogDataAddValueType(builder, Value.ArrayString)


_map_scalar_type_to_serialiser = {
    np.dtype("byte"): _serialise_byte,
    np.dtype("ubyte"): _serialise_ubyte,
    np.dtype("int8"): _serialise_byte,
    np.dtype("int16"): _serialise_short,
    np.dtype("int32"): _serialise_int,
    np.dtype("int64"): _serialise_long,
    np.dtype("uint8"): _serialise_ubyte,
    np.dtype("uint16"): _serialise_ushort,
    np.dtype("uint32"): _serialise_uint,
    np.dtype("uint64"): _serialise_ulong,
    np.dtype("float32"): _serialise_float,
    np.dtype("float64"): _serialise_double,
}

_map_array_type_to_serialiser = {
    np.dtype("byte"): _serialise_bytearray,
    np.dtype("ubyte"): _serialise_ubytearray,
    np.dtype("int8"): _serialise_bytearray,
    np.dtype("int16"): _serialise_shortarray,
    np.dtype("int32"): _serialise_intarray,
    np.dtype("int64"): _serialise_longarray,
    np.dtype("uint8"): _serialise_ubytearray,
    np.dtype("uint16"): _serialise_ushortarray,
    np.dtype("uint32"): _serialise_uintarray,
    np.dtype("uint64"): _serialise_ulongarray,
    np.dtype("float32"): _serialise_floatarray,
    np.dtype("float64"): _serialise_doublearray,
}


def serialise_f142(
    value: Any,
    source_name: str,
    timestamp_unix_ns: int = 0,
    alarm_status: Union[int, None] = None,
    alarm_severity: Union[int, None] = None,
) -> bytes:
    """
    Serialise value and corresponding timestamp as an f142 Flatbuffer message.
    Should automagically use a sensible type for value in the message, but if
    in doubt pass value in as a numpy ndarray of a carefully chosen dtype.

    :param value: only scalar value currently supported; if ndarray then ndim must be 0
    :param source_name: name of the data source
    :param timestamp_unix_ns: timestamp corresponding to value, e.g. when value was measured, in nanoseconds
    :param alarm_status: EPICS alarm status, best to provide using enum-like class defined in logdata_f142.AlarmStatus
    :param alarm_severity: EPICS alarm severity, best to provide using enum-like class defined in logdata_f142.AlarmSeverity
    """
    builder, source = _setup_builder(source_name)
    value = np.array(value)

    if value.ndim == 0:
        _serialise_value(
            builder, source, value, _serialise_string, _map_scalar_type_to_serialiser
        )
    elif value.ndim == 1:
        _serialise_value(
            builder,
            source,
            value,
            _serialise_stringarray,
            _map_array_type_to_serialiser,
        )
    else:
        raise NotImplementedError("f142 only supports scalars or 1D array values")

    return bytes(
        _complete_buffer(builder, timestamp_unix_ns, alarm_status, alarm_severity)
    )


def _serialise_value(
    builder: flatbuffers.Builder,
    source: int,
    value: Any,
    string_serialiser: Callable,
    serialisers_map: Dict,
):
    # We can use a dictionary to map most numpy types to one of the types defined in the flatbuffer schema
    # but we have to handle strings separately as there are many subtypes
    if np.issubdtype(value.dtype, np.unicode_) or np.issubdtype(
        value.dtype, np.string_
    ):
        string_serialiser(builder, value, source)
    else:
        try:
            serialisers_map[value.dtype](builder, value, source)
        except KeyError:
            # There are a few numpy types we don't try to handle, for example complex numbers
            raise NotImplementedError(
                f"Cannot serialise data of type {value.dtype}, must use one of "
                f"{list(_map_scalar_type_to_serialiser.keys()).append(np.unicode_)}"
            )


_map_fb_enum_to_type = {
    Value.Byte: Byte,
    Value.UByte: UByte,
    Value.Short: Short,
    Value.UShort: UShort,
    Value.Int: Int,
    Value.UInt: UInt,
    Value.Long: Long,
    Value.ULong: ULong,
    Value.Float: Float,
    Value.Double: Double,
    Value.String: String,
    Value.ArrayByte: ArrayByte,
    Value.ArrayUByte: ArrayUByte,
    Value.ArrayShort: ArrayShort,
    Value.ArrayUShort: ArrayUShort,
    Value.ArrayInt: ArrayInt,
    Value.ArrayUInt: ArrayUInt,
    Value.ArrayLong: ArrayLong,
    Value.ArrayULong: ArrayULong,
    Value.ArrayFloat: ArrayFloat,
    Value.ArrayDouble: ArrayDouble,
    Value.ArrayString: ArrayString,
}


LogDataInfo = namedtuple(
    "LogDataInfo",
    ("value", "source_name", "timestamp_unix_ns", "alarm_status", "alarm_severity"),
)


def _decode_if_scalar_string(value: np.ndarray) -> Union[str, np.ndarray]:
    if value.ndim == 0 and (
        np.issubdtype(value.dtype, np.unicode_)
        or np.issubdtype(value.dtype, np.string_)
    ):
        return value.item().decode()
    return value


def deserialise_f142(buffer: Union[bytearray, bytes]) -> LogDataInfo:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    log_data = LogData.LogData.GetRootAsLogData(buffer, 0)
    source_name = log_data.SourceName() if log_data.SourceName() else b""

    value_offset = log_data.Value()
    value_fb = _map_fb_enum_to_type[log_data.ValueType()]()
    value_fb.Init(value_offset.Bytes, value_offset.Pos)
    try:
        value = value_fb.ValueAsNumpy()
    except AttributeError:
        try:
            # Must be a scalar value then, so we'll get it like this
            value = np.array(value_fb.Value())
        except TypeError:
            # In that case it is an array of strings, which for some reason doesn't get a generated ValueAsNumpy method
            # So we'll have to extract each element from the buffer manually and construct our own numpy array
            value = np.array(
                [str(value_fb.Value(n), "utf-8") for n in range(value_fb.ValueLength())]
            )

    value = _decode_if_scalar_string(value)

    timestamp = log_data.Timestamp()

    return LogDataInfo(
        value, source_name.decode(), timestamp, log_data.Status(), log_data.Severity()
    )



import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.run_stop_6s4t import deserialise_6s4t, serialise_6s4t


class TestSerialisation6s4t:
    original_entry = {
        "job_id": "some_key",
        "stop_time": 578214,
        "run_name": "test_run",
        "service_id": "filewriter1",
        "command_id": "some command id",
    }

    def test_serialises_and_deserialises_6s4t_message_correctly(self):
        buf = serialise_6s4t(**self.original_entry)
        deserialised_tuple = deserialise_6s4t(buf)

        assert deserialised_tuple.job_id == self.original_entry["job_id"]
        assert deserialised_tuple.stop_time == self.original_entry["stop_time"]
        assert deserialised_tuple.run_name == self.original_entry["run_name"]
        assert deserialised_tuple.service_id == self.original_entry["service_id"]
        assert deserialised_tuple.command_id == self.original_entry["command_id"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_6s4t(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_6s4t(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "6s4t" in SERIALISERS
        assert "6s4t" in DESERIALISERS



from typing import Any, Union, NamedTuple
from collections import namedtuple
import flatbuffers
import numpy as np
from datetime import datetime, timezone

from streaming_data_types.fbschemas.epics_scalar_data_scal import ScalarData
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayInt8 import (
    ArrayInt8,
    ArrayInt8Start,
    ArrayInt8AddValue,
    ArrayInt8End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayUInt8 import (
    ArrayUInt8,
    ArrayUInt8Start,
    ArrayUInt8AddValue,
    ArrayUInt8End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayInt16 import (
    ArrayInt16,
    ArrayInt16Start,
    ArrayInt16AddValue,
    ArrayInt16End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayUInt16 import (
    ArrayUInt16,
    ArrayUInt16Start,
    ArrayUInt16AddValue,
    ArrayUInt16End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayInt32 import (
    ArrayInt32,
    ArrayInt32Start,
    ArrayInt32AddValue,
    ArrayInt32End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayUInt32 import (
    ArrayUInt32,
    ArrayUInt32Start,
    ArrayUInt32AddValue,
    ArrayUInt32End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayInt64 import (
    ArrayInt64,
    ArrayInt64Start,
    ArrayInt64AddValue,
    ArrayInt64End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayUInt64 import (
    ArrayUInt64,
    ArrayUInt64Start,
    ArrayUInt64AddValue,
    ArrayUInt64End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayFloat32 import (
    ArrayFloat32,
    ArrayFloat32Start,
    ArrayFloat32AddValue,
    ArrayFloat32End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.ArrayFloat64 import (
    ArrayFloat64,
    ArrayFloat64Start,
    ArrayFloat64AddValue,
    ArrayFloat64End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.Int8 import Int8
from streaming_data_types.fbschemas.epics_scalar_data_scal.UInt8 import UInt8
from streaming_data_types.fbschemas.epics_scalar_data_scal.Int16 import Int16
from streaming_data_types.fbschemas.epics_scalar_data_scal.UInt16 import UInt16
from streaming_data_types.fbschemas.epics_scalar_data_scal.Int32 import Int32
from streaming_data_types.fbschemas.epics_scalar_data_scal.UInt32 import UInt32
from streaming_data_types.fbschemas.epics_scalar_data_scal.UInt64 import UInt64
from streaming_data_types.fbschemas.epics_scalar_data_scal.Float32 import Float32
from streaming_data_types.fbschemas.epics_scalar_data_scal.Float64 import (
    Float64,
    Float64Start,
    Float64AddValue,
    Float64End,
)
from streaming_data_types.fbschemas.epics_scalar_data_scal.Int64 import (
    Int64,
    Int64Start,
    Int64AddValue,
    Int64End,
)

from streaming_data_types.fbschemas.epics_scalar_data_scal.Value import Value
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"scal"

SerialiserFunctions = namedtuple(
    "SerialiserFunctionMap",
    ("StartFunction", "AddValueFunction", "EndFunction", "value_type_enum"),
)


def _serialise_value(
    builder: flatbuffers.Builder, value: Any, function_map: SerialiserFunctions
):
    function_map.StartFunction(builder)
    function_map.AddValueFunction(builder, value)
    return function_map.EndFunction(builder)


_map_array_type_to_serialiser = {
    np.dtype("int8"): SerialiserFunctions(
        ArrayInt8Start, ArrayInt8AddValue, ArrayInt8End, Value.ArrayInt8
    ),
    np.dtype("int16"): SerialiserFunctions(
        ArrayInt16Start, ArrayInt16AddValue, ArrayInt16End, Value.ArrayInt16
    ),
    np.dtype("int32"): SerialiserFunctions(
        ArrayInt32Start, ArrayInt32AddValue, ArrayInt32End, Value.ArrayInt32
    ),
    np.dtype("int64"): SerialiserFunctions(
        ArrayInt64Start, ArrayInt64AddValue, ArrayInt64End, Value.ArrayInt64
    ),
    np.dtype("uint8"): SerialiserFunctions(
        ArrayUInt8Start, ArrayUInt8AddValue, ArrayUInt8End, Value.ArrayUInt8
    ),
    np.dtype("uint16"): SerialiserFunctions(
        ArrayUInt16Start, ArrayUInt16AddValue, ArrayUInt16End, Value.ArrayUInt16
    ),
    np.dtype("uint32"): SerialiserFunctions(
        ArrayUInt32Start, ArrayUInt32AddValue, ArrayUInt32End, Value.ArrayUInt32
    ),
    np.dtype("uint64"): SerialiserFunctions(
        ArrayUInt64Start, ArrayUInt64AddValue, ArrayUInt64End, Value.ArrayUInt64
    ),
    np.dtype("float32"): SerialiserFunctions(
        ArrayFloat32Start, ArrayFloat32AddValue, ArrayFloat32End, Value.ArrayFloat32
    ),
    np.dtype("float64"): SerialiserFunctions(
        ArrayFloat64Start, ArrayFloat64AddValue, ArrayFloat64End, Value.ArrayFloat64
    ),
}

DoubleFuncMap = SerialiserFunctions(
    Float64Start, Float64AddValue, Float64End, Value.Float64
)
IntFuncMap = SerialiserFunctions(Int64Start, Int64AddValue, Int64End, Value.Int64)


def serialise_scal(
    source_name: str,
    value: Any,
    timestamp: datetime,
) -> bytes:
    builder = flatbuffers.Builder(1024)
    source_name_offset = builder.CreateString(source_name)
    if isinstance(value, int):
        value_offset = _serialise_value(builder, value, IntFuncMap)
        value_type = Value.Int64
    elif isinstance(value, float):
        value_offset = _serialise_value(builder, value, DoubleFuncMap)
        value_type = Value.Float64
    elif isinstance(value, np.ndarray):
        c_func_map = _map_array_type_to_serialiser[value.dtype]
        value_offset = _serialise_value(
            builder, builder.CreateNumpyVector(value), c_func_map
        )
        value_type = c_func_map.value_type_enum
    else:
        raise NotImplementedError(
            f'scal flatbuffer does not support values of type "{type(value)}".'
        )
    ScalarData.ScalarDataStart(builder)
    ScalarData.ScalarDataAddSourceName(builder, source_name_offset)
    ScalarData.ScalarDataAddValue(builder, value_offset)
    ScalarData.ScalarDataAddValueType(builder, value_type)
    ScalarData.ScalarDataAddTimestamp(builder, int(timestamp.timestamp() * 1e9))

    end = ScalarData.ScalarDataEnd(builder)
    builder.Finish(end, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


_map_fb_enum_to_type = {
    Value.Int8: Int8,
    Value.UInt8: UInt8,
    Value.Int16: Int16,
    Value.UInt16: UInt16,
    Value.Int32: Int32,
    Value.UInt32: UInt32,
    Value.Int64: Int64,
    Value.UInt64: UInt64,
    Value.Float32: Float32,
    Value.Float64: Float64,
    Value.ArrayInt8: ArrayInt8,
    Value.ArrayUInt8: ArrayUInt8,
    Value.ArrayInt16: ArrayInt16,
    Value.ArrayUInt16: ArrayUInt16,
    Value.ArrayInt32: ArrayInt32,
    Value.ArrayUInt32: ArrayUInt32,
    Value.ArrayInt64: ArrayInt64,
    Value.ArrayUInt64: ArrayUInt64,
    Value.ArrayFloat32: ArrayFloat32,
    Value.ArrayFloat64: ArrayFloat64,
}


ExtractedScalarData = NamedTuple(
    "ScalarData",
    (
        ("source_name", str),
        ("value", Any),
        ("timestamp", datetime),
    ),
)


def deserialise_scal(buffer: Union[bytearray, bytes]) -> ExtractedScalarData:
    check_schema_identifier(buffer, FILE_IDENTIFIER)
    scalar_data = ScalarData.ScalarData.GetRootAs(buffer, 0)
    source_name = scalar_data.SourceName() if scalar_data.SourceName() else b""

    value_offset = scalar_data.Value()
    value_fb = _map_fb_enum_to_type[scalar_data.ValueType()]()
    value_fb.Init(value_offset.Bytes, value_offset.Pos)
    if hasattr(value_fb, "ValueAsNumpy"):
        value = value_fb.ValueAsNumpy()
    else:
        value = value_fb.Value()
    return ExtractedScalarData(
        source_name=source_name.decode(),
        value=value,
        timestamp=datetime.fromtimestamp(
            scalar_data.Timestamp() / 1e9, tz=timezone.utc
        ),
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class EventMessage(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsEventMessage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = EventMessage()
        x.Init(buf, n + offset)
        return x

    # EventMessage
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # EventMessage
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # EventMessage
    def MessageId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EventMessage
    def PulseTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # EventMessage
    def TimeOfFlight(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # EventMessage
    def TimeOfFlightAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # EventMessage
    def TimeOfFlightLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EventMessage
    def DetectorId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # EventMessage
    def DetectorIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # EventMessage
    def DetectorIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EventMessage
    def FacilitySpecificDataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # EventMessage
    def FacilitySpecificData(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None


def EventMessageStart(builder):
    builder.StartObject(7)


def EventMessageAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def EventMessageAddMessageId(builder, messageId):
    builder.PrependUint64Slot(1, messageId, 0)


def EventMessageAddPulseTime(builder, pulseTime):
    builder.PrependUint64Slot(2, pulseTime, 0)


def EventMessageAddTimeOfFlight(builder, timeOfFlight):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(timeOfFlight), 0
    )


def EventMessageStartTimeOfFlightVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def EventMessageAddDetectorId(builder, detectorId):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(detectorId), 0
    )


def EventMessageStartDetectorIdVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def EventMessageAddFacilitySpecificDataType(builder, facilitySpecificDataType):
    builder.PrependUint8Slot(5, facilitySpecificDataType, 0)


def EventMessageAddFacilitySpecificData(builder, facilitySpecificData):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(facilitySpecificData), 0
    )


def EventMessageEnd(builder):
    return builder.EndObject()



from datetime import datetime
from typing import NamedTuple, Union

import flatbuffers

from streaming_data_types.fbschemas.run_stop_6s4t import RunStop
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"6s4t"


def serialise_6s4t(
    job_id: str,
    run_name: str = "test_run",
    service_id: str = "",
    command_id: str = "",
    stop_time: Union[int, datetime, None] = None,
) -> bytes:
    builder = flatbuffers.Builder(500)
    builder.ForceDefaults(True)

    if service_id is None:
        service_id = ""
    if type(stop_time) is datetime:
        stop_time = int(stop_time.timestamp() * 1000)
    elif stop_time is None:
        stop_time = 0

    service_id_offset = builder.CreateString(service_id)
    job_id_offset = builder.CreateString(job_id)
    run_name_offset = builder.CreateString(run_name)
    command_id_offset = builder.CreateString(command_id)

    # Build the actual buffer
    RunStop.RunStopStart(builder)
    RunStop.RunStopAddServiceId(builder, service_id_offset)
    RunStop.RunStopAddJobId(builder, job_id_offset)
    RunStop.RunStopAddRunName(builder, run_name_offset)
    RunStop.RunStopAddStopTime(builder, stop_time)
    RunStop.RunStopAddCommandId(builder, command_id_offset)

    run_stop_message = RunStop.RunStopEnd(builder)
    builder.Finish(run_stop_message, file_identifier=FILE_IDENTIFIER)

    return bytes(builder.Output())


RunStopInfo = NamedTuple(
    "RunStopInfo",
    (
        ("stop_time", int),
        ("run_name", str),
        ("job_id", str),
        ("service_id", str),
        ("command_id", str),
    ),
)


def deserialise_6s4t(buffer: Union[bytearray, bytes]) -> RunStopInfo:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    run_stop = RunStop.RunStop.GetRootAsRunStop(buffer, 0)
    service_id = run_stop.ServiceId() if run_stop.ServiceId() else b""
    job_id = run_stop.JobId() if run_stop.JobId() else b""
    run_name = run_stop.RunName() if run_stop.RunName() else b""
    stop_time = run_stop.StopTime()
    command_id = run_stop.CommandId()

    return RunStopInfo(
        stop_time=stop_time,
        run_name=run_name.decode(),
        job_id=job_id.decode(),
        service_id=service_id.decode(),
        command_id=command_id.decode(),
    )



from typing import NamedTuple, Optional, Union

import flatbuffers

from streaming_data_types.fbschemas.finished_writing_wrdn import FinishedWriting
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"wrdn"


def serialise_wrdn(
    service_id: str,
    job_id: str,
    error_encountered: bool,
    file_name: str,
    metadata: Optional[str] = None,
    message: Optional[str] = None,
) -> bytes:
    builder = flatbuffers.Builder(500)
    builder.ForceDefaults(True)

    service_id_offset = builder.CreateString(service_id)
    job_id_offset = builder.CreateString(job_id)
    file_name_offset = builder.CreateString(file_name)
    if metadata is not None:
        metadata_offset = builder.CreateString(metadata)
    if message is not None:
        message_offset = builder.CreateString(message)

    # Build the actual buffer
    FinishedWriting.FinishedWritingStart(builder)
    FinishedWriting.FinishedWritingAddServiceId(builder, service_id_offset)
    FinishedWriting.FinishedWritingAddJobId(builder, job_id_offset)
    FinishedWriting.FinishedWritingAddErrorEncountered(builder, error_encountered)
    FinishedWriting.FinishedWritingAddFileName(builder, file_name_offset)
    if metadata:
        FinishedWriting.FinishedWritingAddMetadata(builder, metadata_offset)
    if message:
        FinishedWriting.FinishedWritingAddMessage(builder, message_offset)

    finished_writing_message = FinishedWriting.FinishedWritingEnd(builder)

    builder.Finish(finished_writing_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


WritingFinished = NamedTuple(
    "FinishedWriting",
    (
        ("service_id", str),
        ("job_id", str),
        ("error_encountered", bool),
        ("file_name", str),
        ("metadata", Optional[str]),
        ("message", Optional[str]),
    ),
)


def deserialise_wrdn(buffer: Union[bytearray, bytes]) -> FinishedWriting:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    finished_writing = FinishedWriting.FinishedWriting.GetRootAsFinishedWriting(
        buffer, 0
    )
    service_id = finished_writing.ServiceId()
    job_id = finished_writing.JobId()
    has_error = finished_writing.ErrorEncountered()
    file_name = finished_writing.FileName() if finished_writing.FileName() else b""
    metadata = (
        finished_writing.Metadata().decode() if finished_writing.Metadata() else None
    )
    message = (
        finished_writing.Message().decode() if finished_writing.Message() else None
    )

    return WritingFinished(
        service_id=service_id.decode(),
        job_id=job_id.decode(),
        error_encountered=has_error,
        file_name=file_name.decode(),
        metadata=metadata,
        message=message,
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ArrayInt8(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ArrayInt8()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsArrayInt8(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def ArrayInt8BufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x73\x63\x61\x6C", size_prefixed=size_prefixed
        )

    # ArrayInt8
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ArrayInt8
    def Value(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Int8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # ArrayInt8
    def ValueAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int8Flags, o)
        return 0

    # ArrayInt8
    def ValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ArrayInt8
    def ValueIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0


def ArrayInt8Start(builder):
    builder.StartObject(1)


def Start(builder):
    return ArrayInt8Start(builder)


def ArrayInt8AddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def AddValue(builder, value):
    return ArrayInt8AddValue(builder, value)


def ArrayInt8StartValueVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartValueVector(builder, numElems):
    return ArrayInt8StartValueVector(builder, numElems)


def ArrayInt8End(builder):
    return builder.EndObject()


def End(builder):
    return ArrayInt8End(builder)



import os

from setuptools import find_packages, setup

from streaming_data_types._version import version

DESCRIPTION = "Python utilities for handling ESS streamed data"

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
try:
    with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except Exception as error:
    print("COULD NOT GET LONG DESC: {}".format(error))
    LONG_DESCRIPTION = DESCRIPTION

setup(
    name="ess_streaming_data_types",
    version=version,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="ScreamingUdder",
    url="https://github.com/ess-dmsc/python-streaming-data-types",
    license="BSD 2-Clause License",
    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.6.0",
    install_requires=["flatbuffers==1.12", "numpy"],
    extras_require={"dev": ["flake8", "pre-commit", "pytest", "tox"]},
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FB_Tables

import flatbuffers


class NDAttribute(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsNDAttribute(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = NDAttribute()
        x.Init(buf, n + offset)
        return x

    # NDAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # NDAttribute
    def PName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # NDAttribute
    def PDescription(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # NDAttribute
    def PSource(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # NDAttribute
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # NDAttribute
    def PData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # NDAttribute
    def PDataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # NDAttribute
    def PDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0


def NDAttributeStart(builder):
    builder.StartObject(5)


def NDAttributeAddPName(builder, pName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(pName), 0
    )


def NDAttributeAddPDescription(builder, pDescription):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(pDescription), 0
    )


def NDAttributeAddPSource(builder, pSource):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(pSource), 0
    )


def NDAttributeAddDataType(builder, dataType):
    builder.PrependInt8Slot(3, dataType, 0)


def NDAttributeAddPData(builder, pData):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(pData), 0
    )


def NDAttributeStartPDataVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def NDAttributeEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class String(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsString(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = String()
        x.Init(buf, n + offset)
        return x

    # String
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # String
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def StringStart(builder):
    builder.StartObject(1)


def StringAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def StringEnd(builder):
    return builder.EndObject()



import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.fbschemas.logdata_f142.AlarmSeverity import AlarmSeverity
from streaming_data_types.fbschemas.logdata_f142.AlarmStatus import AlarmStatus
from streaming_data_types.logdata_f142 import deserialise_f142, serialise_f142


class TestSerialisationF142:
    original_entry = {
        "source_name": "some_source",
        "value": 578214,
        "timestamp_unix_ns": 1585332414000000000,
    }

    def test_serialises_and_deserialises_integer_f142_message_correctly(self):
        buf = serialise_f142(**self.original_entry)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == self.original_entry["source_name"]
        assert deserialised_tuple.value == self.original_entry["value"]
        assert (
            deserialised_tuple.timestamp_unix_ns
            == self.original_entry["timestamp_unix_ns"]
        )

    def test_serialises_and_deserialises_byte_f142_message_correctly(self):
        byte_log = {
            "source_name": "some_source",
            "value": 0x7F,
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**byte_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == byte_log["source_name"]
        assert deserialised_tuple.value == byte_log["value"]
        assert deserialised_tuple.timestamp_unix_ns == byte_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_float_f142_message_correctly(self):
        float_log = {
            "source_name": "some_source",
            "value": 1.234,
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**float_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == float_log["source_name"]
        assert deserialised_tuple.value == float_log["value"]
        assert deserialised_tuple.timestamp_unix_ns == float_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_scalar_ndarray_f142_message_correctly(self):
        numpy_log = {
            "source_name": "some_source",
            "value": np.array(42),
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**numpy_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == numpy_log["source_name"]
        assert deserialised_tuple.value == np.array(numpy_log["value"])
        assert deserialised_tuple.timestamp_unix_ns == numpy_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_string_f142_message_correctly(self):
        string_log = {
            "source_name": "some_source",
            "value": "some_string",
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**string_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == string_log["source_name"]
        assert deserialised_tuple.value == string_log["value"]
        assert deserialised_tuple.timestamp_unix_ns == string_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_native_list_correctly(self):
        list_log = {
            "source_name": "some_source",
            "value": [1, 2, 3],
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**list_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == list_log["source_name"]
        # Array values are output as numpy array
        assert np.array_equal(deserialised_tuple.value, np.array(list_log["value"]))
        assert deserialised_tuple.timestamp_unix_ns == list_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_numpy_array_integers_correctly(self):
        array_log = {
            "source_name": "some_source",
            "value": np.array([1, 2, 3]),
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**array_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == array_log["source_name"]
        assert np.array_equal(deserialised_tuple.value, array_log["value"])
        assert deserialised_tuple.timestamp_unix_ns == array_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_numpy_array_preserves_byte_type_correctly(
        self,
    ):
        array_log = {
            "source_name": "some_source",
            "value": np.array([1, 2, 3], dtype=np.uint8),
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**array_log)
        deserialised_tuple = deserialise_f142(buf)

        assert np.array_equal(deserialised_tuple.value, array_log["value"])
        assert deserialised_tuple.value.dtype == array_log["value"].dtype

    def test_serialises_and_deserialises_numpy_array_preserves_integer_type_correctly(
        self,
    ):
        array_log = {
            "source_name": "some_source",
            "value": np.array([1, 2, 3], dtype=np.uint16),
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**array_log)
        deserialised_tuple = deserialise_f142(buf)

        assert np.array_equal(deserialised_tuple.value, array_log["value"])
        assert deserialised_tuple.value.dtype == array_log["value"].dtype

    def test_serialises_and_deserialises_numpy_array_floats_correctly(self):
        array_log = {
            "source_name": "some_source",
            "value": np.array([1.1, 2.2, 3.3]),
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**array_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == array_log["source_name"]
        assert np.allclose(deserialised_tuple.value, array_log["value"])
        assert deserialised_tuple.timestamp_unix_ns == array_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_numpy_array_strings_correctly(self):
        array_log = {
            "source_name": "some_source",
            "value": np.array(["1", "2", "3"]),
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**array_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.source_name == array_log["source_name"]
        assert np.array_equal(deserialised_tuple.value, array_log["value"])
        assert deserialised_tuple.timestamp_unix_ns == array_log["timestamp_unix_ns"]

    def test_serialises_and_deserialises_epics_alarms_correctly(self):
        float_log = {
            "source_name": "some_source",
            "value": 1.234,
            "timestamp_unix_ns": 1585332414000000000,
            "alarm_status": AlarmStatus.HIHI,
            "alarm_severity": AlarmSeverity.MAJOR,
        }
        buf = serialise_f142(**float_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.alarm_status == float_log["alarm_status"]
        assert deserialised_tuple.alarm_severity == float_log["alarm_severity"]

    def test_epics_alarms_default_to_no_change_when_not_provided_to_serialiser(self):
        float_log = {
            "source_name": "some_source",
            "value": 1.234,
            "timestamp_unix_ns": 1585332414000000000,
        }
        buf = serialise_f142(**float_log)
        deserialised_tuple = deserialise_f142(buf)

        assert deserialised_tuple.alarm_status == AlarmStatus.NO_CHANGE
        assert deserialised_tuple.alarm_severity == AlarmSeverity.NO_CHANGE

    def test_raises_not_implemented_error_when_trying_to_serialise_numpy_complex_number_type(
        self,
    ):
        complex_log = {
            "source_name": "some_source",
            "value": np.complex(3, 4),
            "timestamp_unix_ns": 1585332414000000000,
        }
        with pytest.raises(NotImplementedError):
            serialise_f142(**complex_log)

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_f142(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_f142(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "f142" in SERIALISERS
        assert "f142" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class AlarmState(object):
    NONE = 0
    DEVICE = 1
    DRIVER = 2
    RECORD = 3
    DB = 4
    CONF = 5
    UNDEFINED = 6
    CLIENT = 7



from collections import namedtuple
from typing import Optional, Union

import flatbuffers

from streaming_data_types.fbschemas.epics_connection_info_ep00 import (
    EpicsConnectionInfo,
    EventType,
)
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"ep00"


def serialise_ep00(
    timestamp_ns: int,
    event_type: EventType,
    source_name: str,
    service_id: Optional[str] = None,
) -> bytes:
    builder = flatbuffers.Builder(136)
    builder.ForceDefaults(True)

    if service_id is not None:
        service_id_offset = builder.CreateString(service_id)
    source_name_offset = builder.CreateString(source_name)

    EpicsConnectionInfo.EpicsConnectionInfoStart(builder)
    if service_id is not None:
        EpicsConnectionInfo.EpicsConnectionInfoAddServiceId(builder, service_id_offset)
    EpicsConnectionInfo.EpicsConnectionInfoAddSourceName(builder, source_name_offset)
    EpicsConnectionInfo.EpicsConnectionInfoAddType(builder, event_type)
    EpicsConnectionInfo.EpicsConnectionInfoAddTimestamp(builder, timestamp_ns)

    end = EpicsConnectionInfo.EpicsConnectionInfoEnd(builder)
    builder.Finish(end, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


EpicsConnection = namedtuple(
    "EpicsConnection", ("timestamp", "type", "source_name", "service_id")
)


def deserialise_ep00(buffer: Union[bytearray, bytes]) -> EpicsConnection:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    epics_connection = (
        EpicsConnectionInfo.EpicsConnectionInfo.GetRootAsEpicsConnectionInfo(buffer, 0)
    )

    source_name = (
        epics_connection.SourceName() if epics_connection.SourceName() else b""
    )
    service_id = epics_connection.ServiceId() if epics_connection.ServiceId() else b""

    return EpicsConnection(
        epics_connection.Timestamp(),
        epics_connection.Type(),
        source_name.decode(),
        service_id.decode(),
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class AlarmStatus(object):
    NO_ALARM = 0
    READ = 1
    WRITE = 2
    HIHI = 3
    HIGH = 4
    LOLO = 5
    LOW = 6
    STATE = 7
    COS = 8
    COMM = 9
    TIMED = 10
    HWLIMIT = 11
    CALC = 12
    SCAN = 13
    LINK = 14
    SOFT = 15
    BAD_SUB = 16
    UDF = 17
    DISABLE = 18
    SIMM = 19
    READ_ACCESS = 20
    WRITE_ACCESS = 21
    NO_CHANGE = 22



import time
from collections import namedtuple
from typing import Union

import flatbuffers
import numpy as np

from streaming_data_types.fbschemas.NDAr_NDArray_schema import NDArray
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"NDAr"


def serialise_ndar(
    id: str,
    dims: list,
    data_type: int,
    data: list,
) -> bytes:
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    # Build dims
    NDArray.NDArrayStartDimsVector(builder, len(dims))
    # FlatBuffers builds arrays backwards
    for s in reversed(dims):
        builder.PrependUint64(s)
    dims_offset = builder.EndVector(len(dims))

    # Build data
    NDArray.NDArrayStartPDataVector(builder, len(data))
    # FlatBuffers builds arrays backwards
    for s in reversed(data):
        builder.PrependUint8(s)
    data_offset = builder.EndVector(len(data))

    # Build the actual buffer
    NDArray.NDArrayStart(builder)
    NDArray.NDArrayAddDataType(builder, data_type)
    NDArray.NDArrayAddDims(builder, dims_offset)
    NDArray.NDArrayAddId(builder, id)
    NDArray.NDArrayAddPData(builder, data_offset)
    NDArray.NDArrayAddTimeStamp(builder, int(time.time() * 1000))
    nd_array_message = NDArray.NDArrayEnd(builder)

    builder.Finish(nd_array_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


nd_Array = namedtuple(
    "NDArray",
    (
        "id",
        "timestamp",
        "data",
    ),
)


def get_data(fb_arr):
    """
    Converts the data array into the correct type.
    """
    raw_data = fb_arr.PDataAsNumpy()
    numpy_arr_type = [
        np.int8,
        np.uint8,
        np.int16,
        np.uint16,
        np.int32,
        np.uint32,
        np.int64,
        np.uint64,
        np.float32,
        np.float64,
    ]
    return raw_data.view(numpy_arr_type[fb_arr.DataType()]).reshape(
        fb_arr.DimsAsNumpy()
    )


def deserialise_ndar(buffer: Union[bytearray, bytes]) -> NDArray:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    nd_array = NDArray.NDArray.GetRootAsNDArray(buffer, 0)
    id = nd_array.Id()
    timestamp = nd_array.TimeStamp()
    data = get_data(nd_array)

    return nd_Array(
        id=id,
        timestamp=timestamp,
        data=data,
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ScalarData(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ScalarData()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsScalarData(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def ScalarDataBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x73\x63\x61\x6C", size_prefixed=size_prefixed
        )

    # ScalarData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ScalarData
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ScalarData
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # ScalarData
    def ValueType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # ScalarData
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None


def ScalarDataStart(builder):
    builder.StartObject(4)


def Start(builder):
    return ScalarDataStart(builder)


def ScalarDataAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def AddSourceName(builder, sourceName):
    return ScalarDataAddSourceName(builder, sourceName)


def ScalarDataAddTimestamp(builder, timestamp):
    builder.PrependInt64Slot(1, timestamp, 0)


def AddTimestamp(builder, timestamp):
    return ScalarDataAddTimestamp(builder, timestamp)


def ScalarDataAddValueType(builder, valueType):
    builder.PrependUint8Slot(2, valueType, 0)


def AddValueType(builder, valueType):
    return ScalarDataAddValueType(builder, valueType)


def ScalarDataAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def AddValue(builder, value):
    return ScalarDataAddValue(builder, value)


def ScalarDataEnd(builder):
    return builder.EndObject()


def End(builder):
    return ScalarDataEnd(builder)
